
Require Import Arith.
Require Import List.
Require Import Tactics.


Parameter tconstant : Set.
Parameter constant : Set.


Inductive head : Set :=
| const : constant -> head
| var : nat -> head
.


Inductive pattern : Set :=
| shiftm : pattern
| unitpat : pattern
| pairpat : pattern -> pattern -> pattern
.


Inductive sort : Set :=
| re
| vl
| sp
| tm
| nt
| pt
| ts
| kd
| cn
| jd
.


Lemma eqdec_sort :
  forall (z z' : sort), {z = z'} + {z <> z'}.
Proof.
intros z z'.
destruct z; destruct z'; try (left; reflexivity); right; intro H; discriminate H.
Qed.

Hint Resolve eqdec_sort : eqdec.


Inductive exp : sort -> Set :=

(* results *)
| ret : exp vl -> exp re
| use : head -> exp sp -> exp re

(* values *)
| thunk : exp tm -> exp vl
| unitp : exp vl
| pairp : exp vl -> exp vl -> exp vl

(* spines *)
| nl : exp sp
| pm : pattern -> exp re -> exp sp
| app : exp vl -> exp sp -> exp sp
| pi1 : exp sp -> exp sp
| pi2 : exp sp -> exp sp

(* terms *)
| etam : exp re -> exp tm
| shiftp : exp re -> exp tm
| lam : pattern -> exp tm -> exp tm
| pairm : exp tm -> exp tm -> exp tm

(* negative types *)
| tuse : tconstant -> exp ts -> exp nt
| up : exp pt -> exp nt
| pi : pattern -> exp pt -> exp nt -> exp nt
| wth : exp nt -> exp nt -> exp nt
| sigma : exp nt -> exp nt -> exp nt

(* positive types *)
| down : exp nt -> exp pt
| one : exp pt
| tens : exp pt -> exp pt -> exp pt
| ex : pattern -> exp pt -> exp pt -> exp pt

(* type spine *)
| tnl : exp ts
| tapp : exp vl -> exp ts -> exp ts

(* kinds *)
| ty : exp kd
| tpi : pattern -> exp pt -> exp kd -> exp kd

(* consequents *)
| trp : exp pt -> exp cn
| susp : exp nt -> exp cn

(* judgement *)
| judr : exp cn -> exp jd
| judv : exp pt -> exp jd
| juds : exp nt -> exp cn -> exp jd
| judt : exp nt -> exp jd
| judts : exp kd -> exp kd -> exp jd
| judnt : exp jd
| judpt : exp jd
| judk : exp jd
.


Definition res := exp re.
Definition val := exp vl.
Definition spine := exp sp.
Definition term := exp tm.
Definition ntp := exp nt.
Definition ptp := exp pt.
Definition tspine := exp ts.
Definition kind := exp kd.
Definition con := exp cn.
Definition judgement := exp jd.


Definition context := list (exp nt).


Definition sh_var (i j k : nat) : nat :=
  if lt_dec k i then
    k
  else
    j + k.


Definition sh_head (i j : nat) (h : head) : head :=
  (match h with
   | const c => const c
   | var k => var (sh_var i j k)
   end).


Fixpoint bound (p : pattern) {struct p} : nat :=
  (match p with
   | shiftm => 1
   | unitpat => 0
   | pairpat p q => bound p + bound q
   end).


Fixpoint sh {z} (i j : nat) (e : exp z) {struct e} : exp z :=
  (match e with
   | ret v => ret (sh i j v)
   | use h s => use (sh_head i j h) (sh i j s)
   | thunk m => thunk (sh i j m)
   | unitp => unitp
   | pairp m n  => pairp (sh i j m) (sh i j n)
   | nl => nl
   | pm p r => pm p (sh (bound p + i) j r)
   | app v s => app (sh i j v) (sh i j s)
   | pi1 s => pi1 (sh i j s)
   | pi2 s => pi2 (sh i j s)
   | etam r => etam (sh i j r)
   | shiftp r => shiftp (sh i j r)
   | lam p m => lam p (sh (bound p + i) j m)
   | pairm m n => pairm (sh i j m) (sh i j n)
   | tuse c s => tuse c (sh i j s)
   | up a => up (sh i j a)
   | pi p a b => pi p (sh i j a) (sh (bound p + i) j b)
   | wth a b => wth (sh i j a) (sh i j b)
   | sigma a b => sigma (sh i j a) (sh (1 + i) j b)
   | down a => down (sh i j a)
   | one => one
   | tens a b => tens (sh i j a) (sh i j b)
   | ex p a b => ex p (sh i j a) (sh (bound p + i) j b)
   | tnl => tnl
   | tapp v s => tapp (sh i j v) (sh i j s)
   | ty => ty
   | tpi p a k => tpi p (sh i j a) (sh (bound p + i) j k)
   | trp a => trp (sh i j a)
   | susp a => susp (sh i j a)
   | judr a => judr (sh i j a)
   | judv a => judv (sh i j a)
   | juds a u => juds (sh i j a) (sh i j u)
   | judt a => judt (sh i j a)
   | judts a u => judts (sh i j a) (sh i j u)
   | judnt => judnt
   | judpt => judpt
   | judk => judk
   end).


Fixpoint shctx (i j : nat) (G : context) {struct G} : context :=
  (match G with
   | nil => nil
   | A :: G' => sh (i + length G') j A :: shctx i j G'
   end).
